// LIBRARIES ////////////////////////////////////////////////////////////
#include "TimerOne.h"
#include "TimerThree.h"
/////////////////////////////////////////////////////////////////////////


// PIN DEFINITION ///////////////////////////////////////////////////////
//Optointerruptor sensor (interrupt pin)
int blackwheel = 2;
//Ultrasound (interrupt pin for echo)
int echo = 3; //green cable
int trig = 4; //blue cable
//IR line follower sensors
int IRr = 5;
int IRl = 6;
//Buzzer
int buzzer = 7;
//Motor A
int enA = 8;
int in1 = 9;
int in2 = 10;
//Motor B
int in3 = 11;
int in4 = 12;
int enB = 13;
/////////////////////////////////////////////////////////////////////////


// VARIABLES ////////////////////////////////////////////////////////////
//Needed for the counter
unsigned int counter = 0;             //initialize the counter
int oldValue = 0;                     //needed to count completed pulse changes.

//Needed to get the speed
const float diskslots = 20.00;        //number of slots from the black wheel
const float wheeldiameter = 6.61;     //cm
const float pi = 3.14;

//Needed to compare and store speeds
int speedRobot = 0;
int speedObstacle = 0;

//Needed to control the ultrasound
volatile int distance;
const int maxDistance = 50;           //range value [0,400]
const int minDistance = 7;

//Needed to keep track of the echo pulse duration
volatile long echostart = 0;          //Records start of echo pulse  
volatile long echoend = 0;            //Records end of echo pulse         
volatile long echoduration = 0;       //Stores the duration of the echo (difference between end and start)               

//Needed to work with collected data
volatile int speedAuto = 0;           //Stores the speed of the robot    
volatile int distanceA = 0;           //Stores the old distance to the object
volatile int distanceB = 0;           //Stores the new distance to the object

//Needed to determine the car's response
volatile int mode = 0;                //Needed keep the mode updated.
boolean overtakingON = false;         //Needed to activate the overtaking mode.
boolean oncomingON = false;           //NOT USED. Needed to activate the oncoming mode (when a car comes in the oppositte direction).
/////////////////////////////////////////////////////////////////////////



/////////////////////////////////////////////////////////////////////////
// INTERRUPTIONS ////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////

/*
 * INTERRUPTION 0
 * This interruption consist on two functions: 
 * - A counter which counts how many times the sensor input changes during a second.
 * This counter will help us to get the number of spins that the wheel does per second.
 * - A speed calculation function that gets the speed of the auto taking into consideration
 * the rotation (wheel spins) and the lenght of the wheel.
 */

//Set up a counter.
//The sensor input is 1 when light is detected and 0 when there is no light.
//The oldValue variable helps to count a completed pulse change.
void count() {
  /*Serial.println ((int)digitalRead(blackwheel));
  Serial.print (" -- ");
  Serial.println (counter);*/
  if ((int)digitalRead(blackwheel) != oldValue){
    if ((int)digitalRead(blackwheel) == 1){
      counter++;
    }
    oldValue = (int)digitalRead(blackwheel);
  }
}

//Calculate car speed. 
//First we get the number of spins that the wheel does in a second.
//Then we calculate 
void speedISR() {
  Timer1.detachInterrupt();                           //stop the timer
  float rotation = counter / diskslots;               //get number of rotations a.k.a. RPS (Revolutions Per Second).
  float lenght = wheeldiameter * pi;                  //wheel lenght equals the distance run per rotation
  float f_currentSpeed = (rotation * lenght) + 0.5;   //get speed. Sum 0.5 to round the number before converting to integer.
  int currentSpeed = (int) f_currentSpeed;            //convert speed to integer 
  speedAuto = currentSpeed;                           //update global variable
  /*Serial.println(counter);
  Serial.println(rotation);
  Serial.println(lenght);
  Serial.print("Speed: "); 
  Serial.print(speedAuto);
  Serial.println(" cm/s");*/
  counter = 0;                                        //reset counter to 0
  Timer1.attachInterrupt(speedISR);                   //enable the timer
}

/*
 * INTERRUPTION 1
 * This interruption consist on three functions: 
 * - echointerrupt() keeps track of the time between the trigger and the echo.
 * - A function that sends a clean trigger shot each second.
 * - One last function used to update the mode.
 */

//This function is called everytime the echo pin undergoes a change (H to L or L to H).
//From L to H: record starting time.
//From H to L: record ending time.
void echointerrupt(){
  switch (digitalRead(echo)){                     
    case HIGH:                                      
      echoend = 0;                                //reset the end time                                 
      echostart = micros();                       //records the starting time                   
      break;
    case LOW:                                       
      echoend = micros();                         //records the ending time                          
      echoduration = echoend - echostart;         //calculates the echo duration (time the echo pin has kept HIGH)            
      distance = echoduration/58;                 //calculates the distance to the object | distance=(traveldistance/2)*speed of sound
      distanceA = distanceB;
      distanceB = distance;
      /*Serial.print("Distance: ");
      Serial.println(distance);
      Serial.print("DistanceA: ");
      Serial.println(distanceA);
      Serial.print("DistanceB: ");
      Serial.println(distanceB);*/
      updatemode();
      break;
  }
}

//Triggers an ultrasound wave for 10us necessary to initialize measurement.
void distanceISR() {
  //Timer3.detachInterrupt();
  digitalWrite(trig,LOW);
  delayMicroseconds(2);
  digitalWrite(trig,HIGH);
  delayMicroseconds(10);
  digitalWrite(trig,LOW);
  delayMicroseconds(2);
  //Timer3.attachInterrupt(distanceISR);
}

//With this fuction a mode is set up and the system knows the level of risk.
//max & min distances are constants declared at the beginning of the program.
void updatemode() {
  if(distance>maxDistance){ 
    mode = 0;                         //keep running - no danger
  }else{ 
    if(distance>minDistance){ 
      mode = 1;                       //pay atention - sensitive situation
    }else{ 
      mode = 2;                       //stop - imminent collision
    }
  }  
               
}
/////////////////////////////////////////////////////////////////////////



/////////////////////////////////////////////////////////////////////////
// ARDUINO SETUP FUNCTION ///////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////

void setup()
{
  //initialize serial port
  Serial.begin(9600);

  //Seting up interruptions
  //Timer 1
  Timer1.initialize(1000000);                                             //Sets timer for 1s
  attachInterrupt(digitalPinToInterrupt (blackwheel), count, CHANGE);     //Increases counter when speed sensor pin goes HIGH
  Timer1.attachInterrupt(speedISR);                                       //Runs the speedISR function each time the timer period finishes. 
  //Timer3
  Timer3.initialize(1000000);                                             //Sets timer for 1s
  attachInterrupt(digitalPinToInterrupt (echo), echointerrupt, CHANGE);   //starts/pauses the clock when there is a change in the echo pin
  Timer3.attachInterrupt(distanceISR,5000);                               //Triggers a clean shot every half-second.
  
  //Set all the motor control pins to outputs
  //Right
  pinMode(enA, OUTPUT);
  pinMode(in1, OUTPUT);
  pinMode(in2, OUTPUT);
  //Left
  pinMode(enB, OUTPUT);
  pinMode(in3, OUTPUT);
  pinMode(in4, OUTPUT);

  //set Buzzer pin to output
  pinMode(buzzer, OUTPUT);
  
  //Set all the IR sensor pins to inputs
  //Right
  pinMode(IRr, INPUT);
  //Left
  pinMode(IRl, INPUT);


  //set Ultrasound pins to input/output
  //Echo (interruption pin)
  pinMode(echo, INPUT);
  //Trigger
  pinMode(trig, OUTPUT);

  //set Optointerruptor pin to input (interruption pin)
  pinMode(blackwheel, INPUT);
}
/////////////////////////////////////////////////////////////////////////



/////////////////////////////////////////////////////////////////////////
// ARDUINO LOOP FUNCTION ////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////

void loop()
{
  //Sets up the speed of the vehicle. 
  //Range from 0 to 255 - Speed is set to its maximum.
  analogWrite(enA, 255);
  analogWrite(enB, 255); 

  //MODE 0 - Nothing happens. Keep moving.
  while (mode == 0){
    carmove();                          //The car moves normally
  }

  //MODE 1 - Taking control of the situation.
  while (mode == 1){
   compareSpeed();                      //Compares speeds from the robot and obstacle. Decides what to do.
    if (overtakingON == true){          //If overtaking is activated, the car overtakes at the distance of 30cm.
      if (distanceB<30){
        overtaking();
      }
    }else{
      carmove();                        //The car moves normally if overtaking is not activated.
    }
  }

  //MODE 2 - Imminent collision.
  while (mode == 2){
    moveStop();                          //The car stops
  }
}
/////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////
/////////// FUNCTIONS ///////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////

//Basic movements: stop, forward, backwards (not used), turn right & turn left.
void moveStop(){
  //turn off motor A
  digitalWrite(in1, LOW);
  digitalWrite(in2, LOW);
  //turn off motor B
  digitalWrite(in3, LOW);
  digitalWrite(in4, LOW); 
}
void moveForward(){
  //turn on motor A
  digitalWrite(in1, HIGH);
  digitalWrite(in2, LOW);
  //turn on motor B
  digitalWrite(in3, LOW);
  digitalWrite(in4, HIGH);
}
void moveBackwards(){
  //turn on motor A
  digitalWrite(in1, LOW);
  digitalWrite(in2, HIGH);
  //turn on motor B
  digitalWrite(in3, HIGH);
  digitalWrite(in4, LOW);
}
void moveRight(){
  //turn on motor A
  digitalWrite(in1, LOW);
  digitalWrite(in2, LOW);
  //turn on motor B
  digitalWrite(in3, LOW);
  digitalWrite(in4, HIGH); 
}
void moveLeft(){
  //turn on motor A
  digitalWrite(in1, HIGH);
  digitalWrite(in2, LOW);
  //turn on motor B
  digitalWrite(in3, LOW);
  digitalWrite(in4, LOW); 
}
/////////////////////////////////////////////////////////////////////////

//Defines the normal movement of the car and keeps the car on a black line.
//[Line follower implementation]
void carmove()
{
  //START. Both sensors black.
  if ((digitalRead(IRr) == LOW) && (digitalRead(IRl) == LOW)){moveForward();}
  //RIGHT. Correct to right.
  if ((digitalRead(IRr) == LOW) && (digitalRead(IRl) == HIGH)){moveRight();}
  //LEFT. Correct to left.
  if ((digitalRead(IRr) == HIGH) && (digitalRead(IRl) == LOW)){moveLeft();}
  
  //END. Both sensors white. 
  if ((digitalRead(IRr) == HIGH) && (digitalRead(IRl) == HIGH)){moveStop();}
}
/////////////////////////////////////////////////////////////////////////

//Calculates the speed of the obstacle
void getSpeed()
{  
  speedRobot = speedAuto;                                       //Takes the current speed
  speedObstacle = speedRobot + (distanceB - distanceA);         //Gets the speed of the obstacle taking the robot's speed as reference
  
  /*Serial.print("Distancies actuals: ");
  Serial.print(distanceA);
  Serial.print(" i ");
  Serial.println(distanceB);
  Serial.print("Velocitat Arduino: ");
  Serial.println(speedA);
  Serial.print("Velocitat obstacle: ");
  Serial.println(speedObstacle);*/
}  
/////////////////////////////////////////////////////////////////////////

//Compare speeds to determine different scenarios
void compareSpeed() 
{
  getSpeed();                                     //distanceA, distanceB, speedRobot & speedObstacle are updated

  //Negative speed, a vehicle is coming towards us --> ONCOMING
  //Implemented scenario but not tested.
  if (speedObstacle < 0){
    overtakingON = false;
    oncomingON = true;                            //Need to change to another lane. To the left is always better.
  }
  
  //Positive speed
  if (speedObstacle > 0){
    if (speedObstacle >= speedRobot){             //Detected vehicle will eventually get out from our distance limit.
      overtakingON = false;
      oncomingON = false;
    }
    if (speedRobot > speedObstacle){              //Our speed is higher than the detected vehicle's speed.
      if((speedRobot*2/3) >= speedObstacle){      //Detected vehicle speed is too low --> OVERTAKING
        overtakingON = true;
        oncomingON = false;
      }else{                                      //Detected vehicle speed has a similar speed --> Best option is to REDUCE SPEED.
        overtakingON = true;                      //Should be "false" but it is true for testing purposes.
        oncomingON = false;
        //analogWrite(enA, 255);                  //Ideally here is where speeds are adapted.
        //analogWrite(enB, 255);
        //This instruction is not implemented for this version of the prototype. 
        //overtakingON should be false when implementing this part.
      }
    }
  }

  //Detected vehicle is actually an obstacle (not a moving car) --> OVERTAKING
  if (speedObstacle == 0){
    overtakingON = true;
    oncomingON = false;
  }
  
  /*Serial.print("oncomingON : ");
  Serial.println(oncomingON);
  Serial.print("overtakingON : ");
  Serial.println(overtakingON);*/
}  
/////////////////////////////////////////////////////////////////////////

//Moves to the left lane
void changeLaneLeft()
{
  moveLeft();                                                               //Turn left during almost a second
  delay(900);
  while ((digitalRead(IRr) == HIGH) && (digitalRead(IRl) == HIGH)){         //Move forward while a black line is not detected
    moveForward();
  }  
   while ((digitalRead(IRl) == LOW)){                                       //Move forward until the left sensor gets out of the black line
    moveForward();
  }  
   while ((digitalRead(IRl) == HIGH)){                                      //Move right until the left sensor gets in the black line again
    moveRight();
  }  
  moveForward();                                                            //Slight movement to redirect
  delay(200);
}
/////////////////////////////////////////////////////////////////////////

//Moves to the right lane
void changeLaneRight()
{
  moveRight();                                                               
  delay(900);
  while ((digitalRead(IRr) == HIGH) && (digitalRead(IRl) == HIGH)){         
    moveForward();
  }  
   while ((digitalRead(IRr) == LOW)){                                       
    moveForward();
  }  
   while ((digitalRead(IRr) == HIGH)){                                      
    moveLeft();
  }  
  moveForward();
  delay(200);
}
/////////////////////////////////////////////////////////////////////////

//OVERTAKE 
//First changes to the left lane and then goes back to the right lane.
void overtaking(){    
  changeLaneLeft();
  changeLaneRight();                                                        //Same movement on the opossite direction
  overtakingON = false;                                                     //End of OVERTAKING
}
/////////////////////////////////////////////////////////////////////////



/*
/////////////////////////////////////////////////////////////////////////
//////// NOT USED FUNCTIONS /////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////

//Another function to detect an obstacle, usinig the Ping library
#include <NewPing.h>
NewPing sonar(trig, echo, maxDistance);
void getDistance() //Measuring front distance
{
  delay(50); //wait 50ms between pings for a clear shot
  distance = sonar.ping_cm(); //get value in cm
  Serial.print(distance);
  Serial.println("cm");
}
/////////////////////////////////////////////////////////////////////////

//Defines the time needed to turn the car
int getTurnTime(float angle)
{
  speedRobot = speedAuto;
  float rad = angle * pi / 180.00;
  int temps = (rad * 15 / speedRobot) * 1000;
  return temps;
}
/////////////////////////////////////////////////////////////////////////

//Activation of an auditive signal.
void buzzerSound(){
  digitalWrite(buzzer,HIGH);
  delay(1);//wait for 1ms
  digitalWrite(buzzer,LOW);
  delay(1);//wait for 1ms
}
/////////////////////////////////////////////////////////////////////////

//Another function to overtake
void overtaking2(){
  //1700 - 600
  moveLeft();
  delay(1000);
  while ((digitalRead(IRr) == HIGH) && (digitalRead(IRl) == HIGH)){
    //Serial.println("FORWARD");
    //Serial.println(digitalRead(IRr));
    //Serial.println(digitalRead(IRl));
    moveForward();
  }  
  //Serial.println("OUT Bucle 1");
  moveRight();
  delay(700);
  while ((digitalRead(IRr) == HIGH) && (digitalRead(IRl) == HIGH)){
    moveForward();
  }
  //moveForward();
  //delay(1000);
  carmove();  
  overtakingON = false;
}
/////////////////////////////////////////////////////////////////////////

//ONCOMING
//Change to the right lane (and stays)
void oncoming(){
  changeLaneRight(); 
  oncomingON = false; 
}
/////////////////////////////////////////////////////////////////////////
*/
